# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------
import sys
import sphinx_gallery


project = "vfi"
copyright = "2019, Christos Aridas"
author = "Christos Aridas"

# The full version, including alpha/beta/rc tags
release = "0.0.1"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "numpydoc",
    "sphinx_gallery.gen_gallery",
]

sphinx_gallery_conf = {
    "examples_dirs": "../../examples",  # path to your example scripts
    "gallery_dirs": "auto_examples",  # path to where to save gallery generated output
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

autosummary_generate = True
plot_gallery = True

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
master_doc = "index"
version = "0.0.1"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]


# intersphinx configuration
intersphinx_mapping = {
    "python": ("https://docs.python.org/{.major}".format(sys.version_info), None),
    "numpy": ("https://docs.scipy.org/doc/numpy/", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/reference", None),
    "matplotlib": ("https://matplotlib.org/", None),
    "sklearn": ("http://scikit-learn.org/stable", None),
}



import numbers

import numpy as np
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.preprocessing import LabelEncoder, KBinsDiscretizer, normalize
from sklearn.utils.multiclass import check_classification_targets
from sklearn.utils.validation import check_X_y, check_array, check_is_fitted


class VFI(BaseEstimator, ClassifierMixin):
    """ Classification by voting feature intervals.

    Intervals are constucted around each class for each attribute
    (basically discretization). Class counts are recorded for each
    interval on each attribute. Classification is by voting.

    Parameters
    ----------
    n_bins : int or array-like, shape (n_features,) (default="auto")
        The number of bins to produce. When is set to 'auto' the n_bins
        equals to the double of the number of classes. Raises ValueError
        if n_bins < 2.

    strategy : {'uniform', 'quantile', 'kmeans'}, (default='quantile')
        Strategy used to define the widths of the bins.

        uniform
            All bins in each feature have identical widths.
        quantile
            All bins in each feature have the same number of points.
        kmeans
            Values in each bin have the same nearest center of a 1D k-means
            cluster.


    Attributes
    ----------
    classes_ : array, shape (n_classes,)
        The classes.

    classes_distribution_ : array, shape (n_classes,)
        The distribution of the classes.

    interval_class_counts_ : array, shape (n_features, n_bins, n_classes,)
        Contains the raw class counts per feature and per bin.

    n_bins_ : int or array-like, shape (n_features,).
        The number of bins used during fit.

    n_classes_ : int
        The number of classes.


    References
    ----------
    .. [1] G. Demiroz, A. Guvenir: Classification by voting feature intervals.
           In: 9th European Conference on Machine Learning, 85-92, 1997.01.

    """

    def __init__(self, n_bins="auto", strategy="uniform"):
        self.n_bins = n_bins
        self.strategy = strategy

    def fit(self, X, y):
        """ Fit VFI according to X, y.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The training input samples.
        y : array-like, shape (n_samples,)
            The target values. An array of int.

        Returns
        -------
        self : object
               Returns self.
        """

        self._validate_params()
        X, y = check_X_y(X, y)
        check_classification_targets(y)

        self._label_encoder = LabelEncoder()
        y_transformed = self._label_encoder.fit_transform(y)
        self.classes_ = self._label_encoder.classes_
        self.n_classes_ = len(self.classes_)
        n_features = X.shape[1]
        self.class_distribution_ = np.bincount(y_transformed, minlength=self.n_classes_)

        self.n_bins_ = self.n_classes_ * 2 if self.n_bins == "auto" else self.n_bins

        discretizer = KBinsDiscretizer(
            n_bins=self.n_bins_, encode="ordinal", strategy=self.strategy
        )
        X_discretized = discretizer.fit_transform(X).astype(np.int32)
        self.interval_class_counts_ = np.zeros(
            (n_features, self.n_bins_, self.n_classes_)
        )

        for i, col in enumerate(X_discretized.T):
            for bin in range(discretizer.n_bins_[i]):
                class_counts = np.bincount(
                    y_transformed[col == bin], minlength=self.n_classes_
                )
                self.interval_class_counts_[i, bin] = class_counts

        self._discretizer = discretizer

        return self

    def predict(self, X):
        """ Perform classification on an array of test vectors X.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The input samples.

        Returns
        -------
        y : ndarray, shape (n_samples,)
            The label for each sample is the label of the closest sample
            seen during fit.
        """

        check_is_fitted(self, ["classes_"])
        X = check_array(X)

        predicted_probabilitiy = self.predict_proba(X)
        return self.classes_.take((np.argmax(predicted_probabilitiy, axis=1)), axis=0)

    def predict_proba(self, X):
        """ Return probability estimates for the test vector X.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)

        Returns
        -------
        probas : array-like of shape (n_samples, n_classes)
                 Returns the probability of the samples for each class in
                 the model. The columns correspond to the classes in sorted
                 order, as they appear in the attribute :term:`classes_`.
        """

        check_is_fitted(self, ["classes_"])
        X = check_array(X)

        X_discretized = self._discretizer.transform(X).astype(int)
        class_votes = np.zeros((X.shape[0], self.n_classes_))

        for i, col in enumerate(X_discretized.T):
            feature_votes = (
                self.interval_class_counts_[i][col] / self.class_distribution_
            )
            class_votes += normalize(feature_votes, axis=1, norm="l1")

        probas = normalize(class_votes, axis=1, norm="l1")

        return probas

    def _validate_params(self):
        if isinstance(self.n_bins, numbers.Integral) and self.n_bins < 2:
            msg = (
                "{} received an invalid value for parameter 'n_bins'. "
                "Valid values are 'auto' or any integer > 1."
            )
            msg = msg.format(self.__class__.__name__)
            raise ValueError(msg)



#! /usr/bin/env python
"""A template for scikit-learn compatible packages."""

import os

from setuptools import find_packages, setup

ver_file = os.path.join("vfi", "_version.py")
with open(ver_file) as f:
    exec(f.read())

DISTNAME = "vfi"
DESCRIPTION = "Voting by Feature Intervals classifier."
with open("README.rst", encoding="utf-8") as f:
    LONG_DESCRIPTION = f.read()
MAINTAINER = "Christos K. Aridas"
MAINTAINER_EMAIL = "ichkoar@gmail.com"
URL = "https://github.com/chkoar/vfi"
LICENSE = "MIT"
DOWNLOAD_URL = "https://github.com/chkoar/vfi"
VERSION = __version__
INSTALL_REQUIRES = ["numpy", "scipy", "scikit-learn"]
CLASSIFIERS = [
    "Intended Audience :: Science/Research",
    "Intended Audience :: Developers",
    "License :: OSI Approved",
    "Programming Language :: Python",
    "Topic :: Software Development",
    "Topic :: Scientific/Engineering",
    "Operating System :: Microsoft :: Windows",
    "Operating System :: POSIX",
    "Operating System :: Unix",
    "Operating System :: MacOS",
    "Programming Language :: Python :: 2.7",
    "Programming Language :: Python :: 3.5",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
]
EXTRAS_REQUIRE = {
    "tests": ["black", "flake8", "invoke", "pytest", "pytest-cov"],
    "docs": [
        "invoke",
        "matplotlib",
        "numpydoc",
        "sphinx",
        "sphinx-gallery",
        "sphinx_rtd_theme",
    ],
}

setup(
    name=DISTNAME,
    maintainer=MAINTAINER,
    maintainer_email=MAINTAINER_EMAIL,
    description=DESCRIPTION,
    license=LICENSE,
    url=URL,
    version=VERSION,
    download_url=DOWNLOAD_URL,
    long_description=LONG_DESCRIPTION,
    zip_safe=False,  # the package can run out of an .egg file
    classifiers=CLASSIFIERS,
    packages=find_packages(),
    install_requires=INSTALL_REQUIRES,
    extras_require=EXTRAS_REQUIRE,
)



"""
===========================
Sensitivity analysis
===========================

In this plot we perform a sensitivity analysis of the n_bins parameter and we can see how the different values affect the performance of the classifier.
"""

print(__doc__)

import matplotlib.pyplot as plt
import numpy as np
import vfi

from sklearn.datasets import load_iris
from sklearn.model_selection import validation_curve

X, y = load_iris(1)

param_range = range(2, 21)
train_scores, test_scores = validation_curve(
    vfi.VFI(),
    X,
    y,
    param_name="n_bins",
    param_range=param_range,
    cv=5,
    scoring="accuracy",
    n_jobs=1,
)
train_scores_mean = np.mean(train_scores, axis=1)
train_scores_std = np.std(train_scores, axis=1)
test_scores_mean = np.mean(test_scores, axis=1)
test_scores_std = np.std(test_scores, axis=1)

plt.title("Validation Curves with VFI")
plt.xlabel("n_bins")
plt.ylabel("Score")
plt.ylim(0.0, 1.0)
plt.xticks(param_range)

lw = 2
plt.plot(
    param_range, train_scores_mean, label="Training score", color="darkorange", lw=lw
)
plt.fill_between(
    param_range,
    train_scores_mean - train_scores_std,
    train_scores_mean + train_scores_std,
    alpha=0.2,
    color="darkorange",
    lw=lw,
)
plt.plot(
    param_range, test_scores_mean, label="Cross-validation score", color="navy", lw=lw
)
plt.fill_between(
    param_range,
    test_scores_mean - test_scores_std,
    test_scores_mean + test_scores_std,
    alpha=0.2,
    color="navy",
    lw=lw,
)
plt.legend(loc="best")
plt.show()



import pytest

from sklearn.datasets import load_iris
from sklearn.utils.estimator_checks import check_estimator

from vfi import VFI


def test_esimator():
    return check_estimator(VFI)


def test_invalid_n_bins():
    X, y = load_iris(return_X_y=True)
    model = VFI(n_bins=1)
    error_message = "Valid values are 'auto' or any integer > 1."
    with pytest.raises(ValueError, match=error_message):
        model.fit(X, y)
